import pandas as pd
import os
import logging
import sys
import pyodbc
from time import time
from io import BytesIO
import boto3

"""
    https://github.com/mkleehammer/pyodbc/wiki/Data-Types


"""

class lynxClient():
    def __init__(self):
        """
            At runtime, environment variables are injected via entrypoint.sh.
            For local development, we fetch environment variables from local enviroment. Make sure they are set up. 
            Lynx only has 1 environment, hence no need to pass ENV
        """
        try: 
            server_raw = os.environ['LYNX_SERVER']
            port = os.environ['LYNX_PORT']
            database = os.environ['LYNX_DATABASE']
            username = os.environ['LYNX_USERNAME']
            password = os.environ['LYNX_PASSWORD']
            server = f'{server_raw},{port}'
            logging.info(f"Fetched lynx credentials from environment variables")
        except Exception as e:
            logging.error("Issue fetching lynx credentials from environment variables. Exiting...")
            logging.error(e)
            sys.exit(1)

        # connect to lynx SQL server 
        try: 
            connection_string = 'DRIVER={ODBC Driver 17 for SQL Server};SERVER=' + server + ';DATABASE=' + database + ';UID=' + username + ';PWD=' + password
            self.conn = self._connect(connection_string)
            logging.info("Succesfully set up connection with Lynx SQL Server")
        except Exception as e:
            logging.error("Unable to establish connection with Lynx SQL Server. Exiting")
            logging.info(e)
            sys.exit(1)
        return

    def _connect(self,connection_string:str ):
        conn = pyodbc.connect(connection_string)
        # https://github.com/mkleehammer/pyodbc/wiki/Unicode
        # conn.setdecoding(pyodbc.SQL_CHAR, encoding='latin1', to=str)
        # conn.setencoding(str, encoding='latin1')
        return conn

    def query_to_list(self, query:str):
        """
            fetchall() where all rows will be stored in memory
            returns 
                data list
                columns list
                dtypes list
        """
        cursor = self.conn.cursor()
        cursor.execute(query)
        rows = cursor.fetchall()
        # https://stackoverflow.com/questions/5504340/python-mysqldb-connection-close-vs-cursor-close
        columns = [column[0] for column in cursor.description]
        dtypes = [column[1] for column in cursor.description]
        cursor.close()
        return rows, columns, dtypes

    def query_fetch_single_value(self, query: str) -> str:
        """
            grab a scalar, eg select max(x) from y
        """
        cursor = self.conn.cursor()
        cursor.execute(query)
        val = cursor.fetchval()
        cursor.close()
        return val

    def query_to_df(self, query:str) -> pd.DataFrame:
        """
            fetchall() where all rows will be stored in memory
            returns pandas dataframe
        """
        cursor = self.conn.cursor()
        cursor.execute(query)
        rows = cursor.fetchall()
        columns = [column[0] for column in cursor.description]
        cursor.close()
        df = pd.DataFrame.from_records(rows, columns=columns)
        return df

    def get_max_cursors(self) -> str:
        cursor = self.conn.cursor()
        nbr_cursors = cursor.getinfo(pyodbc.SQL_MAX_CONCURRENT_ACTIVITIES)
        return nbr_cursors

    def cursor(self):
        cursor = self.conn.cursor()
        return cursor


    def stream_to_s3_parquet(self, query:str, batch_size:int, s3_bucket:str, s3_prefix:str, s3_filename:str) -> None:
        """
            Streams the results of a sql query to parquet files on s3 with 'batch_size' nbr of rows per file. 
            Output files have the following key:
                {s3_bucket}{s3_prefix}{timestamp}_{filename}.parquet        
        """
        for x in self._stream_results(query, batch_size):
            col_names = [column[0] for column in x[0].cursor_description]
            df = pd.DataFrame.from_records(x, columns=col_names)
            self._df_to_parquet_s3(df, s3_bucket, s3_prefix, s3_filename)
        return

                
    def _stream_results(self, query:str, batch_size: int):
        cursor = self.conn.cursor()
        iter = cursor.execute(query)
        while True:
            rows = iter.fetchmany(batch_size)
            if not rows:
                break
            yield rows
            # for row in rows:
            #     yield row

    def _df_to_parquet_s3(self, df:pd.DataFrame, s3_bucket: str, s3_prefix: str, file_name:str):
        parquet_buffer = BytesIO()
        df.to_parquet(parquet_buffer, index=False, allow_truncated_timestamps=True)
        s3 = boto3.resource('s3')
        timestamp = round(time(), 4)
        s3_key = f"{s3_prefix}{timestamp}_{file_name}.parquet"
        s3.Object(s3_bucket, s3_key).put(Body=parquet_buffer.getvalue())
        logging.info(f'Stored {s3_key} on s3 {s3_bucket}')
        del df
        return
    




        


import logging
import sys
import boto3
from io import StringIO
import os

class LogFormatter(logging.Formatter):

    COLOR_CODES = {
        logging.CRITICAL: "\033[1;35m", # bright/bold magenta
        logging.ERROR:    "\033[1;31m", # bright/bold red
        logging.WARNING:  "\033[1;33m", # bright/bold yellow
        logging.INFO:     "\033[0;37m", # white / light gray
        logging.DEBUG:    "\033[1;30m"  # bright/bold black / dark gray
    }

    RESET_CODE = "\033[0m"

    def __init__(self, color, *args, **kwargs):
        super(LogFormatter, self).__init__(*args, **kwargs)
        self.color = color

    def format(self, record, *args, **kwargs):
        if (self.color == True and record.levelno in self.COLOR_CODES):
            record.color_on  = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
        else:
            record.color_on  = ""
            record.color_off = ""
        return super(LogFormatter, self).format(record, *args, **kwargs)


class logger():
    
    def __init__(self, script_name, console_log_output="stdout", console_log_level="info", console_log_color=True, \
            logfile_log_level="debug", logfile_log_color=True ):
        self.script_name = os.path.splitext(os.path.basename(script_name))[0]
        self.console_log_output = console_log_output
        self.console_log_level = console_log_level
        self.console_log_color = console_log_color
        self.logfile_log_level = logfile_log_level
        self.logfile_log_color = logfile_log_color
        self.s3_log = self._setup_logging()
        
    # Setup logging
    def _setup_logging(self):
    # Create logger
    # For simplicity, we use the root logger, i.e. call 'logging.getLogger()'
    # without name argument. This way we can simply use module methods for
    # for logging throughout the script. An alternative would be exporting
    # the logger, i.e. 'global logger; logger = logging.getLogger("<name>")'
        logger = logging.getLogger()
        log_line_template="%(color_on)s[%(created)d] [%(threadName)s] [%(levelname)-8s] %(message)s%(color_off)s"

        # Set global log level to 'debug' (required for handler levels to work)
        logger.setLevel(logging.DEBUG)

        # Create console handler
        console_log_output = self.console_log_output.lower()
        if (console_log_output == "stdout"):
            console_log_output = sys.stdout
        elif (console_log_output == "stderr"):
            console_log_output = sys.stderr
        else:
            print("Failed to set console output: invalid output: '%s'" % console_log_output)
            return False
        console_handler = logging.StreamHandler(console_log_output)

        # Set console log level
        try:
            console_handler.setLevel(self.console_log_level.upper()) # only accepts uppercase level names
        except:
            print("Failed to set console log level: invalid level: '%s'" % self.console_log_level)
            return False

        # Create and set formatter, add console handler to logger
        console_formatter = LogFormatter(fmt=log_line_template, color=self.console_log_color)
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)

        # Create log file handler
        try:
            logfile_handler = logging.FileHandler(f'{self.script_name}.log')
        except Exception as exception:
            print("Failed to set up log file: %s" % str(exception))
            return False

        # Set log file log level
        try:
            logfile_handler.setLevel(self.logfile_log_level.upper()) # only accepts uppercase level names
        except:
            print("Failed to set log file log level: invalid level: '%s'" % self.logfile_log_level)
            return False

        # Create and set formatter, add log file handler to logger
        logfile_formatter = LogFormatter(fmt=log_line_template, color=self.logfile_log_color)
        logfile_handler.setFormatter(logfile_formatter)
        logger.addHandler(logfile_handler)

        # add s3 logging functionality
        s3_log = StringIO()
        s3_log_handler = logging.StreamHandler(s3_log)
        # Set log file log level
        try:
            s3_log_handler.setLevel(self.logfile_log_level.upper()) # only accepts uppercase level names
        except:
            print("Failed to set log file log level: invalid level: '%s'" % self.logfile_log_level)
            return False

        # Create and set formatter, add log file handler to logger
        logfile_formatter = LogFormatter(fmt=log_line_template, color=self.logfile_log_color)
        s3_log_handler.setFormatter(logfile_formatter)
        logger.addHandler(s3_log_handler)

        # boto3 put_object operations result in the entire file's content getting logged as a hex-encoded byte string
        # following line suppreses this behavior
        logging.getLogger("boto3.resources.action").setLevel(logging.INFO)
        # we want to suppres [WARNING ] Connection pool is full, discarding connection, hence we suppres urrllib < ERROR
        logging.getLogger("urllib3").setLevel(logging.ERROR)

        # Success
        return s3_log


class s3Logger():

    def __init__(self, logger, s3_bucket, s3_log_key):
        self.logger = logger.s3_log
        self.s3_bucket = s3_bucket
        self.s3_log_key = s3_log_key

    def send_logs(self):
        s3 = boto3.client("s3")
        s3.put_object(Body=self.logger.getvalue(), Bucket=self.s3_bucket, Key=self.s3_log_key)  


    

    









import psycopg2
import os, sys
import logging
import pandas as pd
import json

class rsClient():
    def __init__(self, env, rs_iam_role):
        """
            When deploying, environment variables are injected depending on ENV via entrypoint.sh.
            For local development, we fetch environemnt variables from local enviroment where we have variables for dev and prd, hence the except statement.
        """
        try:
            dbname=os.environ[f'REDSHIFT_DB']
            user = os.environ[f'REDSHIFT_UNAME']
            password = os.environ[f'REDSHIFT_PASSWORD']
            host = os.environ[f'REDSHIFT_HOST']
            port = os.environ[f'REDSHIFT_PORT']
        except: 
            dbname=os.environ[f'REDSHIFT_DB_{env.upper()}']
            user = os.environ[f'REDSHIFT_UNAME_{env.upper()}']
            password = os.environ[f'REDSHIFT_PASSWORD_{env.upper()}']
            host = os.environ[f'REDSHIFT_HOST_{env.upper()}']
            port = os.environ[f'REDSHIFT_PORT_{env.upper()}']
        try:
            logging.info('Establishing connection to redshift via psycopg2')
            self.conn = psycopg2.connect(dbname=dbname,user = user,password = password,host = host, port = port)
            logging.info('Redshift Connection succesfully established')
        except psycopg2.DatabaseError as error:
            logging.error('Issue connecting to redshift. Exiting with status 1')
            logging.error(error)
            sys.exit(1)
        self.rs_iam_role = rs_iam_role

    def _enable_autocommit(self):
        self.conn.set_session(autocommit=True)
        return

    def _disable_autocommit(self):
        self.conn.set_session(autocommit=False)
        return

    def _query(self, sql):
        with self.conn.cursor() as cur:
            cur.execute(sql)
            self.conn.commit()
        return 

    def _create_stage_like_target(self,rs_target, rs_stage):
        self._query(f"""DROP TABLE IF EXISTS {rs_stage}""")
        self._query(f"""CREATE TABLE {rs_stage} (LIKE {rs_target})""")
        self._enable_autocommit() 
        self._query(f"""ALTER TABLE {rs_stage} DROP COLUMN meta_loading_date_utc""")
        self._disable_autocommit()


    def _copy_parquet_into_stage(self, files, rs_stage):
        for file in files:
            try:
                dml = f"""
                    COPY {rs_stage}
                    FROM '{file}'
                    IAM_ROLE '{self.rs_iam_role}'
                    FORMAT AS PARQUET
                """
                self._query(dml)
            except Exception as e:
                logging.error(f'Issue copying {file} into staging table. Exiting')
                logging.error(f'message: {e}', exc_info=True)
                sys.exit(1)
        logging.info(f'Succesfully loaded all files in temp staging table')
        return

    def _upsert(self, rs_target, rs_stage, upsert_keys):
        where_condition_target = ''
        counter = 0
        for col in upsert_keys:
            if counter == 0:
                line_target = f'{rs_target}.{col} = {rs_stage}.{col}'
            else:
                line_target = f'AND {rs_target}.{col} = {rs_stage}.{col}'
            where_condition_target += str(line_target)
            counter =+ 1
        try:
            self._query(f"""
                    begin transaction;

                    DELETE FROM {rs_target} 
                    USING {rs_stage} 
                    WHERE {where_condition_target} AND
                        GREATEST({rs_stage}.datecreated, {rs_stage}.datemodified) >= GREATEST({rs_target}.datecreated, {rs_target}.datemodified)
                    ;
                    
                    INSERT INTO {rs_target}
                    SELECT *
                    FROM {rs_stage};

                    DROP TABLE {rs_stage};
                    end transaction;
                """)
            logging.info(f'UPSERT succesfull for {rs_stage}')
        except Exception as e:
                logging.error(f'Issue UPSERTING stage into target. Exiting...')
                logging.error(f'message: {e}', exc_info=True)
                sys.exit(1)
    
    def _full_refresh(self, rs_target, rs_stage):
        try:
            self._query(f"""
                    begin transaction;

                    DELETE FROM {rs_target} 
                    ;
                    
                    INSERT INTO {rs_target}
                    SELECT *
                    FROM {rs_stage};

                    DROP TABLE {rs_stage};
                    end transaction;
                """)
            logging.info(f'FULL REFRESH succesfull for {rs_target}')
        except Exception as e:
                logging.error(f'Issue REFRESHING target. Exiting...')
                logging.error(f'message: {e}', exc_info=True)
                sys.exit(1)
    
    
    # TO DO 
    def _append(self):
        None

    def upsert(self, files, rs_target_schema, rs_target_table, upsert_keys):
        """
            ARGUMENTS
                files: list of parquet files (eg. ['s3://bucket/folder/sub/file.parquet', ...])
                rs_target_schema: redshift target schema (eg. 'ingest)
                rs_target_table: redshift target table (eg. 'cogenius_xxx')
                upsert_keys: list of fields to match records between the source (ie. stage) and target, in order to identify which records already exist
                            
            RETURNS
                None

            ADDITIONAL INFO
                List of upsert keys can be considered a composite key. If the composite key already exists within the target table, the associated record
                will be deleted and replaced by the new record with the same composite key. 
                Since and older partition can be reprocessed, an additional check is performed on datemodified. Only if the datemodified within the target is 
                < than the datemodified within the partition we're upserting, the record will be deleted from the target
        """
        rs_target = f'{rs_target_schema}.{rs_target_table}'
        rs_stage = f'{rs_target}_TEMP_STAGE'
        self._create_stage_like_target(rs_target, rs_stage)
        self._copy_parquet_into_stage(files, rs_stage)
        self._upsert(rs_target, rs_stage, upsert_keys)
        return


    def full_refresh(self,files, rs_target_schema, rs_target_table):
        """
            ARGUMENTS
                files: list of parquet files (eg. ['s3://bucket/folder/sub/file.parquet', ...])
                rs_target_schema: redshift target schema (eg. 'ingest)
                rs_target_table: redshift target table (eg. 'cogenius_xxx')
                            
            RETURNS
                None

            ADDITIONAL INFO
                The target table will be truncated (ie we're using delete since truncate commits automatically)
                and stage will be copied into target
        """
        rs_target = f'{rs_target_schema}.{rs_target_table}'
        rs_stage = f'{rs_target}_TEMP_STAGE'
        self._create_stage_like_target(rs_target, rs_stage)
        self._copy_parquet_into_stage(files, rs_stage)
        self._full_refresh(rs_target, rs_stage)
        return


     # Not tested yet
    def load_copy_csv(self, s3_bucket, s3_prefix, rs_target, iam_role, delimiter, columns=None, timeformat='YYYY-MM-DDTHH:MI:SS'):
        """
            s3_bucket: bucket
            s3_prefix: path/path/path/
            rs_target: target_schema.target_table
            env: dev or prd
            columns: ['col1', 'col2', ...]

            ColumnA in your data file will map to col1 in rs, columnB in your file will map to col2 in rs.
            Columns >= len(columns in file). You can use this to create a RS table with X cols, a file with X-1 cols and have 1 col in redshift 
            defaulting to a value, for example a loading date (ie. loading date TIMESTAMP default sysdate)
        """
        path = 's3://'+s3_bucket+'/'+s3_prefix
        if columns == None:
            dml = f"""
                COPY {rs_target}
                FROM '{path}'
                CVS
                IAM_ROLE '{iam_role}'
                delimiter '{delimiter}'
                timeformat '{timeformat}'
                EMPTYASNULL
            """
        else: 
            dml = f"""
                COPY {rs_target} ('{",".join(columns)}')
                FROM '{path}'
                CSV
                IAM_ROLE '{iam_role}'
                delimiter '{delimiter}'
                timeformat '{timeformat}'
                EMPTYASNULL
            """
        self._query(dml)
        return

    def rs_to_df(self, dml):
        df = pd.read_sql_query(dml, self.conn)
        return df

    def rs_column_to_api_query_param_list(self, rs_source_schema:str, rs_source_table:str, rs_source_column:str, api_query_params:str):
        """
            ARGUMENTS:
                rs_source_schema: eg. ingest
                rs_source_table: eg. table
                rs_source_column: eg column1
                api_query_params: eg objectId in case API endpoint is url/objectId?value1 where value1 is extracted from redshift ingest.table.column1
            RETURNS
                list of stringified json [{'objectId': column1[value1]}, {'objectId': column1[value2]}]

            ADDITIONAL INFO:
                For several APIs we need a redshift column as input. For each cell value we'll fetch a data object from an API via query parameters.
                Hence, we create a list of stringified dictionaries which can be readily passed into a fetch function containing the relevant cell value and
                query parameter.
        """
        dml = f"""
            SELECT "{rs_source_column}" from {rs_source_schema}.{rs_source_table} where "{rs_source_column}" is not null
        """
        df = self.rs_to_df(dml)
        # sometimes we have floats in redshift that should be integers, hence we need to cast to integers.
        float_col = df.select_dtypes(include=['float64']) 
        for col in float_col.columns.values:
            df[col] = df[col].astype('int64')
        input_list= df[rs_source_column].tolist()
        api_queries = [json.dumps({api_query_params: x}) for x in input_list]
        return api_queries


    def rs_column_to_list(self, rs_source_schema:str, rs_source_table:str, rs_source_column:str):
        dml = f"""
            SELECT "{rs_source_column}" from {rs_source_schema}.{rs_source_table} where "{rs_source_column}" is not null
        """
        df = self.rs_to_df(dml)
        # sometimes we have floats in redshift that should be integers, hence we need to cast to integers.
        float_col = df.select_dtypes(include=['float64']) 
        for col in float_col.columns.values:
            df[col] = df[col].astype('int64')
        path_params= df[rs_source_column].tolist()
        return path_params

    def rs_query_to_list(self, query: str, rs_source_column: str):
        df = self.rs_to_df(query)
          # sometimes we have floats in redshift that should be integers, hence we need to cast to integers.
        float_col = df.select_dtypes(include=['float64']) 
        for col in float_col.columns.values:
            df[col] = df[col].astype('int64')
        path_params= df[rs_source_column].tolist()
        return path_params




import requests
import json
import random
from time import sleep
import logging
import ast

class basicAPI():

    def __init__(self, url, auth_type = 'api_key', **kwargs ):
        if url.endswith("/"):
            None
        else:
            url = f'{url}/'
        self.API_URL= url
        self.auth_type = auth_type
        self.api_key = kwargs.get('api_key')
        if self.auth_type == 'api_key':
            self.headers = {
                'X-API-KEY' : self.api_key,
                'Accept': 'application/json'
            } 
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        return 

    def fetch_query_params(self, query_params, endpoint, retries=3, backoff_in_seconds=1):
        """
            ARGUMENTS
                query_params: list of stringified dictionary, eg. '{'ean': 12345}'
                endpoint: endpoint part within API endpoint eg. self.url/V1/endpoint?ean=12345

            RETURNS
                response_dictionary or a runtimeError
            
        """
        x = 0
        while True:
            try:
                return self._fetch_query_params(endpoint, query_params)
            except RuntimeError:
                if x == int(retries)-1:
                    raise RuntimeError(f'No 200 response after {retries} tries for query: {json.loads(query_params)}')
                else:
                    logging.warning(f'Trying again ... query: {json.loads(query_params)}')
                    sleep_duration = (int(backoff_in_seconds) * 2 ** x + random.uniform(0, 1))
                    sleep(sleep_duration)
                    x += 1

    def fetch_path_params(self, path_params, endpoint, retries=3, backoff_in_seconds=1):
        x = 0
        while True:
            try:
                return self._fetch_path_params(endpoint, path_params)
            except RuntimeError:
                if x == int(retries)-1:
                    raise RuntimeError(f'No 200 response after {retries} tries for {endpoint}/{path_params}')
                else:
                    logging.warning(f'Trying again ... {endpoint}/{path_params}')
                    sleep_duration = (int(backoff_in_seconds) * 2 ** x + random.uniform(0, 1))
                    sleep(sleep_duration)
                    x += 1

    def _fetch_query_params(self, endpoint, query_params):
        url_endpoint = f'{self.API_URL}{endpoint}'
        response = self.session.get(url_endpoint, params = json.loads(query_params))
        if response.status_code != 200:
            raise RuntimeError(f' {response.status_code} response for {json.loads(query_params)}')
        response_dict=json.loads(response.text)
        # we need to add the query parameter to the response in order to link tables. We add query parameter as api_id
        # for adresmatch, we add straatnaam, huisnr, gemeentenaam en postcode
        d = ast.literal_eval(query_params)
        if len(d)==1:
            (k, v), = d.items()
            response_dict['api_id'] = v
        else:
            for k,v in d.items():
                response_dict[k] = v
        return response_dict

    def _fetch_path_params(self, endpoint, path_params):
        url_endpoint = f'{self.API_URL}{endpoint}/{path_params}'
        response = self.session.get(url_endpoint)
        if response.status_code != 200:
            raise RuntimeError(f' {response.status_code} response for {endpoint}/{path_params}')
        response_dict=json.loads(response.text)
        # we need to add the path parameter to the response
        response_dict['api_id']=path_params
        return response_dict

    def close_session(self):
        try:
            self.session.close()
            logging.info("Closed API session")
        except Exception as e:
            logging.error("Error closing API session")






### WORK IN PROGRESS





import boto3
import pandas as pd
import logging
from io import BytesIO
from time import time
import psutil
import sys
import os
import shutil
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame as SparkDataFrame
import pyspark.sql.functions as F





class sparkClient():

    # file_name = file.split('/')[-1]

    def __init__(self, s3_client, s3_bucket, s3_target_prefix, partition_start, **kwargs):
        self.s3_bucket = s3_bucket
        self.s3_prefix = s3_target_prefix
        self.partition_date = partition_start
        self.s3_client = s3_client
        self.spark = self._create_spark_session()
        return None

    def _create_spark_session(self, loglevel = "ERROR"):
         # https://stackoverflow.com/questions/50891509/apache-spark-codegen-stage-grows-beyond-64-kb 
        nbr_cores = self._get_nbr_cores()
        spark = SparkSession.builder.master(f"local[{nbr_cores}]") \
                    .appName(f'Spark_{self.s3_prefix}') \
                    .config('spark.sql.codegen.wholeStage', 'false') \
                    .config("spark.sql.session.timeZone", "UTC") \
                    .getOrCreate()
        # .config('spark.sql.parquet.filterPushdown', 'false') \
        spark.sparkContext.setLogLevel(loglevel)
        return spark

    def _get_nbr_cores(self) -> int:
       return psutil.cpu_count(logical = False)

    def read_parquet_to_pyspark(self, file:str, delete_file=True) -> SparkDataFrame:
        """
            ARGUMENTS
                file: ./data/filename.parquet
            RETURNS
                spark_df

            ADDITIONAL INFO
                Download file locally with s3_client.download_s3_to_local_file()
                Downloading straight from S3 is complex via pyspark, hence we dowload locally first. 
                Additionally, we delete the local file afterwards if delete_file = True
        """
        spark_df = self.spark.read.parquet(file)

        if delete_file and spark_df:
            os.remove(file)
        return spark_df

    def add_metadata(self, df:SparkDataFrame, file_name: str, partition_date:str) -> SparkDataFrame:
        df = df.withColumn('META_file_name', F.lit(str(file_name))) \
                .withColumn('META_partition_date', F.lit(datetime.strptime(partition_date, "%Y-%m-%d"))) \
                .withColumn('META_processing_date_utc', F.lit(datetime.utcnow()))
        return df

    def convert_int_with_null_back_to_int(self, df: SparkDataFrame, cols: list) -> SparkDataFrame:
        """
            When reading from SQL, columns which are INT but contain NULLS get converted to parquet float columns.
            We create a pyspark schema based on the original SQL schema, hence INT column gets translated into StructField('col', IntegerType(), True).
            In order to apply this pyspark schema, we need to convert the relevant int columns back to int.
        """
        map_to_int = F.udf(lambda x: self._map_float_to_int(x))
        for col in cols:
            df = df.withColumn(col, (map_to_int(df[col])).cast("string"))
        return df


    def _map_float_to_int(self, num):
        if num is None:
            return None
        return str(int(num))


    def spark_apply_schema(self, df:SparkDataFrame , schema) -> SparkDataFrame:
        spark_df = self.spark.createDataFrame(df.collect(), schema = schema)
        return spark_df


import os
import yaml
import inspect
from datetime import date, datetime
import logging
import sys
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
import psutil

cores = psutil.cpu_count(logical = False)

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def get_config(): 
    """
        RETURN
            config object enabling fetching of config from config.yml via dot notation

        ADDITIONAL INFO
            In docker image we copy config file into current dir, for local development we fetch config from level lower
    """
    caller_file = os.path.dirname(inspect.stack()[1].filename)
    try: 
        with open(caller_file + '/config.yml') as file:
            config = yaml.load(file, Loader=yaml.FullLoader)
    except:
        with open(caller_file + '/../config.yml') as file:
            config = yaml.load(file, Loader=yaml.FullLoader)
    config = dotdict(config)
    return config


def create_partition_key(partition_date:str) -> str:
    """
        ARGUMENT
            partition_date: date of airflow task start date. eg 2020-01-01 12:00
        
        RETURN
            year=2020/month=01/day=01/    
    """
    format = "%Y-%m-%d"
    if not isinstance(partition_date, date):
        partition_date = datetime.strptime(partition_date, format)
    day = '{:02d}'.format(partition_date.day)
    month = '{:02d}'.format(partition_date.month)
    year = partition_date.year
    return f"year={year}/month={month}/day={day}/"



def get_args(args):
    """
        ARGUMENTS
            args: sys.args

        RETURNS
            env, pipeline_type, partition_start, partition_start, partition_key

        ADDITIONAL INFO
            Command line args are passed to python programs via airflow dag - AWS Batchoperator - Overrides - Command  
            These args are then passed to our program via entrypoint.sh via $@

            Env, pipeline type and partition start ( ie. airflow task start date) are mandatory for incremental/full pipelines.
            For incremental pipeline, additionally we require a partition end date. 
            Partition key is also returned which is calculated based on partition_start
    """
    try: 
        print(args)
        env = args[1]
        pipeline_type = args[2]
        assert env in ['local', 'dev', 'prd']
        assert pipeline_type in ['full', 'incremental']
        logging.info(f"Environment: {env}")
        logging.info(f"Pipeline type: {pipeline_type}")
    except AssertionError as e:
        logging.error("Incorrect environment or pipeline type argument passed. Exiting...")
        logging.error(e)
        sys.exit(1)
    except Exception:
        logging.error("No environment argument passed. Exiting...")
        sys.exit(1)

    # fetch airflow partition(s)
    try:
        format = "%Y-%m-%d"
        partition_start = args[3]
        try:
            is_valid_start = bool(datetime.strptime(partition_start, format))
        except:
            is_valid_start = False
        assert is_valid_start
        logging.info(f"Partition start: {partition_start}")
        if pipeline_type == 'incremental':
            partition_end = args[4]
            try:
                is_valid_end = bool(datetime.strptime(partition_end, format))
            except:
                is_valid_end= False
            assert is_valid_end
            logging.info(f"Partition end: {partition_end}")
        else:
            partition_end = None
        partition_key = create_partition_key(partition_start)
    except AssertionError as e:
        logging.error("Incorrect partition date argument passed. Dates must be passed as YYYY-MM-DD Exiting...")
        logging.error(e)
        sys.exit(1)
    except: 
        logging.error("No partition date arguments passed. Exiting")
        logging.info("Pass partition start and partition end date arguments as postional arguments 2 and 3")
        sys.exit(1)
    return env, pipeline_type, partition_start, partition_end, partition_key


def chunker(seq, size):
    """
        process list in batches of size 
        eg. for group in chunker(list, 100):
                    print(group)
    """
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


def multithreading_list(func, list, *args, max_workers=cores, **kwargs):
    """
        Multithreading of a function an a input list. Use for heavy IO operations
        If no workers are specified, it will default to the number of processors on the machine
        https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
        Returns 2 dictionaries: 
            filename: dataframes 
            filename: error
    """
    results = {}
    errors= {}
    with ThreadPoolExecutor(max_workers) as executor:
        future_to_x = {executor.submit(func, x, *args, **kwargs): x for x in list}
        for future in as_completed(future_to_x):
            x = future_to_x[future]
            try:
                result = future.result()
                results[x] = result
            except Exception as exc:
                errors[x] = exc
                print('%r generated an exception: %s' % (x, exc))
    return results, errors


from pyspark.sql.types import StructType, StructField, StringType, IntegerType, BooleanType, TimestampType, DoubleType, DecimalType
from pyspark.sql import DataFrame as SparkDataFrame
import pandas as pd

# ensure consistency with transformations.add_metadata
meta_cols = [ ',\n"META_file_name" VARCHAR','"META_partition_date" TIMESTAMP','"META_processing_date_utc" TIMESTAMP','"META_loading_date_utc" TIMESTAMP default sysdate']

spark_meta_cols = {
    'META_file_name': StringType(),
    'META_partition_date': TimestampType(),
    'META_processing_date_utc': TimestampType()
}

spark_to_rs_meta_cols = [',\n"META_loading_date_utc" TIMESTAMP default sysdate']

mapping_sqlserver_redshift = {
    'bigint': 'BIGINT',
    'bigint IDENTITY':'BIGINT',
    'bit':'BOOLEAN',
    'datetime':'TIMESTAMP',
    'datetime2':'TIMESTAMP',
    'decimal': 'NUMERIC',
    'int':'BIGINT',
    'numeric':'NUMERIC',
    'nvarchar': 'VARCHAR',
    # in cogenius, version is captured as timestamp which we arbitrarily translate to varchar
    'timestamp': 'VARCHAR(256)',
    'varchar': 'VARCHAR',
    'uniqueidentifier': 'VARCHAR(256)'
}

mapping_pandas_redshift = {
    'int64': 'BIGINT',
    'float64': 'DOUBLE PRECISION',
    'object': 'VARCHAR(MAX)'
}

mapping_pyspark_redshift = {
    'integer': 'BIGINT',
    'string': 'VARCHAR(MAX)',
    'double': 'DOUBLE PRECISION',
    'timestamp': 'TIMESTAMP',
    'boolean': 'BOOLEAN'
}

mapping_pandas_pyspark = {
    'int64': IntegerType(),
    'float64': DoubleType(),
    'object': StringType(),
    'bool': BooleanType(),
    'datetime64': TimestampType() 
}

mapping_sqlserver_pyspark = {
    'bigint': IntegerType(),
    'bigint IDENTITY': IntegerType(),
    'int': IntegerType(),
    'bit': BooleanType(),
    'datetime':TimestampType(),
    'datetime2': TimestampType(),
    'decimal': DecimalType(),
    'numeric': DecimalType(),
    'nvarchar': StringType(),
    # in cogenius, version is captured as timestamp which we arbitrarily translate to varchar
    'timestamp': StringType(),
    'varchar': StringType(),
    'uniqueidentifier': StringType()
}

def df_to_pyspark(df:pd.DataFrame):
    cols = df.columns
    dtypes = df.dtypes
    target_cols = []
    d = dict(zip(cols, dtypes))
    for key, val in d.items():
        key = key.strip().replace(' ', '_').replace('(','').replace(')','').replace(',','').replace(';','').replace('{','').replace('}','').replace('\n','').replace('\t','').replace('=','')
        target_type = mapping_pandas_pyspark[val.name]
        line = StructField(key, target_type, True)
        target_cols.append(line)
    for key, val in spark_meta_cols.items():
        line = StructField(key, val, True)
        target_cols.append(line)
    pyspark_schema = StructType(target_cols)
    return pyspark_schema

def sql_ddl_to_pyspark(sql_ddl):
    from simple_ddl_parser import DDLParser
    result = DDLParser(sql_ddl).run(output_mode='mssql')
    columns = result[0]['columns']
    cols = []
    dtypes = []
    target_cols = []
    for col in columns:
        cols.append(col['name'])
        dtypes.append(col['type'])
    d = dict(zip(cols, dtypes))
    for key, val in d.items():
        key = key.strip().replace(' ', '_').replace('(','').replace(')','').replace(',','').replace(';','').replace('{','').replace('}','').replace('\n','').replace('\t','').replace('=','')
        target_type = mapping_sqlserver_pyspark[val]
        line = StructField(key, target_type, True)
        target_cols.append(line)
    for key, val in spark_meta_cols.items():
        line = StructField(key, val, True)
        target_cols.append(line)
    pyspark_schema = StructType(target_cols)
    return pyspark_schema
    
def pyspark_to_flyway_ddl(pyspark_schema, target_table, flyway_schema='ingest_schema', ):
    json_sch = pyspark_schema.jsonValue()
    target_cols = []
    for field in json_sch['fields']:
        target_name = field['name']
        target_type = mapping_pyspark_redshift[field['type']]
        line = f'"{target_name}" {target_type}'
        target_cols.append(line)

    joined_cols = ",\n".join(target_cols)
    joined_meta_cols = ",\n".join(spark_to_rs_meta_cols)
    final_cols = joined_cols+joined_meta_cols

    new_ddl = f"""CREATE TABLE IF NOT EXISTS ${{{flyway_schema}}}.{target_table} (
        {final_cols}
    )

    DISTSTYLE AUTO;

    ALTER TABLE ${{{flyway_schema}}}.{target_table} owner to ${{username}};
    """
    return new_ddl

def csv_to_flyway_ddl(cols, dtypes, target_table, flyway_schema='ingest_schema'):
    """
        ARGUMENTS
            cols: df.columns
            dtypes: df.dtypes
        RETURNS 
            flyway ddl 
    """
    target_cols = []
    d = dict(zip(cols, dtypes))
    for key, val in d.items():
        target_type = mapping_pandas_redshift[val.name]
        line = f'"{key}" {target_type}'
        target_cols.append(line)

    joined_cols = ",\n".join(target_cols)
    joined_meta_cols = ",\n".join(meta_cols)
    final_cols = joined_cols+joined_meta_cols

    flyway_ddl = f"""CREATE TABLE IF NOT EXISTS ${{{flyway_schema}}}.{target_table} (
        {final_cols}
    )

    DISTSTYLE AUTO;

    ALTER TABLE ${{{flyway_schema}}}.{target_table} owner to ${{username}};
    """
    return flyway_ddl

## function to convert cogenius sql server DDL to flyway compatible ddl
def sqlserver_to_flyway_ddl(sql_ddl:str, target_table:str, flyway_schema='ingest_schema'):
    """
        ARGUMENTS
            sql_ddl: generated sql ddl from cogenius DB 
            target_table: redshift table name without schema
    """


    from simple_ddl_parser import DDLParser
    result = DDLParser(sql_ddl).run(output_mode='mssql')
    columns = result[0]['columns']
    target_cols = []
    for col in columns:
        source_name = col['name']
        source_type = col['type']
        source_size = col['size']
        target_type = mapping_sqlserver_redshift[source_type]
        if target_type in ['VARCHAR']:
            if source_size:
                None
            else:
                source_size = 'MAX'
            target_type=f'{target_type}({source_size})'
        if target_type in ['NUMERIC']:
            target_type=f'{target_type}{source_size}'
        line = f'"{source_name}" {target_type}'
        target_cols.append(line)

    joined_cols = ",\n".join(target_cols)
    joined_meta_cols = ",\n".join(meta_cols)
    final_cols = joined_cols+joined_meta_cols

    new_ddl = f"""CREATE TABLE IF NOT EXISTS ${{{flyway_schema}}}.{target_table} (
        {final_cols}
    )

    DISTSTYLE AUTO;

    ALTER TABLE ${{{flyway_schema}}}.{target_table} owner to ${{username}};
    """
    return new_ddl

def write_flyway_ddl_to_file(flyway_ddl, target_table, flyway_schema='ingest'):
    from os import listdir
    from os.path import isfile, join

    flyway_dir = (f'/Users/pieterdepetter/GitHub-VEB/terra-etl/flyway/modules/redshift/{flyway_schema}')
    files = [f for f in listdir(flyway_dir) if isfile(join(flyway_dir, f))]
    files.sort()
    version, seq = files[-1].split('__')[0].split('_')
    new_version = int(version.split('V')[-1])+1
    new_version = f'V{new_version:03}'
    new_name = f'{new_version}_1__create_{target_table}.sql'
    with open(f'{flyway_dir}/{new_name}', 'w') as file:
        file.write(flyway_ddl)
    file.close()




import boto3
import pandas as pd
import datetime
import logging
from io import BytesIO
from time import time
import psutil
import sys
import os
import shutil
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame as SparkDataFrame
import pyspark.sql.functions as F
import numpy as np


class tfClient():
    def __init__(self,s3_client, s3_bucket, s3_target_prefix, partition_start, **kwargs):
        self.s3_bucket = s3_bucket
        self.s3_prefix = s3_target_prefix
        self.partition_date = partition_start
        self.s3_client = s3_client
        self.spark = self._create_spark_session()
        return

    def _create_spark_session(self, loglevel = "ERROR"):
         # https://stackoverflow.com/questions/50891509/apache-spark-codegen-stage-grows-beyond-64-kb 
        nbr_cores = self._get_nbr_cores()
        spark = SparkSession.builder.master(f"local[{nbr_cores}]") \
                    .appName(f'Spark_{self.s3_prefix}') \
                    .config('spark.sql.codegen.wholeStage', 'false') \
                    .config("spark.sql.session.timeZone", "UTC") \
                    .getOrCreate()
        spark.sparkContext.setLogLevel(loglevel)
        return spark

    def _get_nbr_cores(self):
       return psutil.cpu_count(logical = False)

    def pandas_read_parquet(self, file, **kwargs) -> pd.DataFrame:
        return pd.read_parquet(file, **kwargs)

    def pandas_read_csv(self, file,  **kwargs) -> pd.DataFrame:
        return pd.read_csv(file, **kwargs)

    ## in case we have null dates, they cannot be 9999-01-01 since this compromises our upsert operation
    def pandas_format_timestamps(self, df:pd.DataFrame) -> pd.DataFrame:
        datetime_df = df.select_dtypes(include="datetime")
        datetime_cols = datetime_df.columns
        # replace cogenius NULL with datetimelike string in order to convert via .dt accessor
        if datetime_cols:
            df[datetime_cols] = df[datetime_cols].replace({None: '1111-01-01 00:00:00.000'})
            df[datetime_cols] = df[datetime_cols].apply(pd.to_datetime, format='%Y-%m-%d %H:%M:%S', errors='coerce')
        return df

    # ensure consistency with parser.meta_cols, parser.spark_meta_cols
    def pandas_add_metadata(self, file, df:pd.DataFrame) -> pd.DataFrame:
        df['META_file_name'] = file
        df['META_partition_date'] = self.partition_date
        df['META_partition_date'] = df['META_partition_date'].apply(pd.to_datetime, format='%Y-%m-%d', errors='coerce')
        df['META_processing_date_utc'] = datetime.datetime.utcnow()
        df['META_processing_date_utc'] = df['META_processing_date_utc'].apply(pd.to_datetime, format='%Y-%m-%d %H:%M:%S', errors='coerce')
        return df

    def pandas_write_to_parquet(self, file, df: pd.DataFrame) -> None:
        file_name = file.split('/')[-1]
        parquet_buffer = BytesIO()
        df.to_parquet(parquet_buffer, index=False, allow_truncated_timestamps=True)
        s3 = boto3.resource('s3')
        # we keep raw file name but overwrite extension if not parquet already
        if '.csv' in file_name:
            file_name = file_name.replace('.csv', '.parquet')
        if '.xlsx' in file_name:
            file_name = file_name.replace('.xslx', '.parquet')
        if '.xls' in file_name:
            file_name = file_name.replace('.xls', '.parquet')
        if '.parquet' in file_name:
            file_name = file_name
        s3_key = f"{self.s3_prefix}{file_name}"
        s3.Object(self.s3_bucket, s3_key).put(Body=parquet_buffer.getvalue())
        logging.info(f'Stored {s3_key} on s3 {self.s3_bucket}')
        del df
        return

    def pandas_convert_objects_to_string(self, df: pd.DataFrame) -> pd.DataFrame:
        columns = df.columns
        for col in columns:
            if (dict(df.dtypes)[col]) in ['object']:
                df[col] = df[col].map(str)
        return df

    def pandas_to_spark_dataframe(self, df:pd.DataFrame, schema) -> SparkDataFrame:
        spark_df = self.spark.createDataFrame(df, schema = schema)
        return spark_df

    def spark_apply_schema(self, df:SparkDataFrame , schema) -> SparkDataFrame:
        spark_df = self.spark.createDataFrame(df.collect(), schema = schema)
        return spark_df

    def spark_empty_and_nan_to_null(self, df: SparkDataFrame) -> SparkDataFrame:
        """
            This function might return columns containing only nulls. This will mess with the column datatypes. Hence, make sure
            you apply spark_apply_schema function after so the null columns are stored in parquet with the correct datatype as expected by redshift
        """
        columns = df.columns
        for column in columns:
            if dict(df.dtypes)[column] in ['int', 'double']:
                df = df.withColumn(column,F.when(F.isnan(F.col(column)),None).otherwise(F.col(column)))
            df = df.withColumn(column, F.when(F.col(column) == '', None).otherwise(F.col(column)))
            df = df.withColumn(column, F.when(F.col(column) == ' ', None).otherwise(F.col(column)))
            df = df.withColumn(column, F.when(F.col(column) == '  ', None).otherwise(F.col(column)))
            if dict(df.dtypes)[column] not in ['int', 'double']:
                df = df.withColumn(column, F.when(F.col(column) == 'NaN', None).otherwise(F.col(column)))
                df = df.withColumn(column, F.when(F.col(column) == 'NaT', None).otherwise(F.col(column)))
        return df

    def spark_write_to_parquet(self, spark_df, max_records_per_file = 10000):
        local_path = './data'
        try:
            shutil.rmtree(local_path)
        except:
            None
        try:
            spark_df.write.option('maxRecordsPerFile', max_records_per_file).mode('overwrite').parquet(local_path)
            for file in os.listdir(local_path):
                local_file = f'{local_path}/{file}'
                if '.crc' in file:
                    os.remove(local_file)
                    continue
                elif 'SUCCESS' in file:
                    os.remove(local_file)
                    continue
                else:
                    self.s3_client.upload_local_file(local_file, self.s3_prefix)
                    os.remove(local_file)
            try:
                os.rmdir(local_path)
            except:
                None
            logging.info(f'Succesfully wrote {self.s3_prefix}')
        except Exception as e:
            logging.error(f'Issue creating parquet file: {self.s3_prefix}. Exiting...')
            logging.error(e)
            sys.exit(1)
        return 

    def pandas_convert_int_null_back_to_int(self, df: pd.DataFrame, columns: list) -> pd.DataFrame:
        for col in columns:
            df[col] = df[col].astype('Int64')
            # convert to None because otherwise we can't load into pyspark IntegerType
            df[col] = df[col].replace({np.nan: None})
        return df



